from enum import Enum
import functools
import warnings


WARNING_PREFIX = "\n\n        Warning:\n            "

MULTIPLE_EXPERIMENTAL_WARNING_MESSAGE = (
    "Multiple features are experimental and may behave unexpectedly:\n\n            * "
)
MULTIPLE_EXPERIMENTAL_WARNING_JOINER = "\n            * "

DEFAULT_EXPERIMENTAL_WARNING_MESSAGE = (
    "This feature is experimental and may behave unexpectedly. "
    "It may also be removed or changed in the future."
)

UPDATE_STEP_WARNING_MESSAGE = (
    "This feature is experimental and may behave unexpectedly. It is your "
    "responsibility to check the state and behavior of the step after an update. "
    "If the creation of the new step configuration takes more than roughly a "
    "minute, this function will return (time out) before the update terminated "
    "and you will need to check the step status manually at regular intervals. "
    "If several step updates are applies at the same time, "
    "the behavior is undefined."
)

PERIODIC_WARNING_MESSAGE = (
    "The periodic deployment execution rule is experimental and may be unstable."
)


class EXPERIMENTAL_VALUES(Enum):
    """Enumeration for experimental_parameters special values."""

    ALL = "ALL"


def experimental(func_or_message):
    """Decorator to mark a function as experimental.

    Args:
        func_or_message (:obj:`str` :obj:`Nothing`): If a string is given, it will be
            used as a warning message. Else if nothing is given to the decorator
            (i.e. `@experimental`), a default warning message will be used.

    """

    def _get_wrapper(func, warning_message):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(warning_message, FutureWarning)
            return func(*args, **kwargs)

        doc_start, doc_separator, doc_end = func.__doc__.partition("\n\n")
        wrapper.__doc__ = (
            doc_start + WARNING_PREFIX + warning_message + doc_separator + doc_end
        )
        return wrapper

    if callable(func_or_message):  # No parameter is given
        func = func_or_message
        return _get_wrapper(func, DEFAULT_EXPERIMENTAL_WARNING_MESSAGE)
    else:  # A message is given
        warning_message = func_or_message
        return lambda func: _get_wrapper(func, warning_message)


def experimental_parameters(parameters):
    """Decorator to mark only some parameters of a function as experimental.

    Args:
        parameters (dict): A dictionary of parameters to be marked as
            experimental. The keys are the names of the parameters and the
            values are dictionaries with the following structure:
            {
                "value1": "message1",
                "value2": "message2",
            }
            where the keys are the values of the parameters and the values are
            the messages to be displayed when the parameter is used.
            You can also use the `EXPERIMENTAL_VALUES.ALL` key to mark all the
            values of a parameter as experimental.

    """

    def actual_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            for key, value in kwargs.items():
                if key in parameters:
                    if value in parameters[key]:
                        warnings.warn(
                            parameters[key][value],
                            FutureWarning,
                        )
                    elif EXPERIMENTAL_VALUES.ALL in parameters[key]:
                        warnings.warn(
                            parameters[key][EXPERIMENTAL_VALUES.ALL],
                            FutureWarning,
                        )
            return func(*args, **kwargs)

        message_list = [
            value for key in parameters for value in parameters[key].values()
        ]
        if len(message_list) > 1:
            warning_message = f"{MULTIPLE_EXPERIMENTAL_WARNING_MESSAGE}\
{MULTIPLE_EXPERIMENTAL_WARNING_JOINER.join(message_list)}"
        else:
            warning_message = message_list[0]
        doc_start, doc_separator, doc_end = func.__doc__.partition("\n\n")
        wrapper.__doc__ = (
            f"{doc_start}{WARNING_PREFIX}{warning_message}{doc_separator}{doc_end}"
        )
        return wrapper

    return actual_decorator
